/* $Workfile: AlvisBurst.c $       $Revision: 11 $      $Modtime: 10/15/10 2:47p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    AlvisBurst.c

    A simple example console application demonstrating the use of the Burst
    device with the following boards:
    
      Alvis-2-CSI 
      Alvis-4-CSI and Alvis-4M-CSI
      Alvis-8-CSI and Alvis-8M-CSI
      Alvis-2-PBX (test configuration; for Odin use only)
      Alvis-4-PBX (test configuration; for Odin use only)
      Alvis-8-PBX (test configuration; for Odin use only)
      Alvis-DMP (hosted by ASM or PCIe)

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.


    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxT1E1.h"
#include "OtxBurst.h"
#include "OtxBurstD.h"
#include "OtxDsp.h"

#ifdef DEMO_UDP_PACKETS
#include "sockets.h"
#endif

#define DEMO_ASP_BURST_RX_FILE_NAME "/tmp/rx.raw"

#define DEMO_DEFAULT_LOG_FILE "AlvisBurst.log"

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

#define DEMO_BURST_BUF_SIZE (4096*2) // 8k

#define LI_PER_BURST 4

#define DEMO_MAX_FRAMES_IN_PACKET 8

#define DEMO_DEFAULT_PORT 8080

//#define DEMO_DEFAULT_DEST "10.0.2.1"//"192.168.102.236"
#define DEMO_DEFAULT_DEST "10.0.1.1"

typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  char                   rxFileName[OTX_MAX_PATH];
  OTX_INT32              nBoardType;
  int                    coLi;
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  OTX_UINT16             nLineLength;
  OTX_BOOL               bReceiveEqualizer;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  OTX_BOOL               bDisableBurstTx;
  OTX_BOOL               bDisableBurstRx;
  OTX_BOOL               bDisableBurstRxFifo;
  OTX_INT32              coBurstHwyTimeslots; // 128 on Alvis-4 and 256 on Alvis-8
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
OTX_EVENT   g_hBurstNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hBurstEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_UINT32  g_nSubsystemId = 0;

typedef struct sBurstStat {
  OTX_UINT32  nRxDataAvailable;
  OTX_UINT32  nTxBufDepleted;
  OTX_UINT32  nSenderOverRunErrors;
  OTX_UINT32  nReceiverUnderRunErrors;
  OTX_UINT32  nWriteDataErrors;
  OTX_UINT32  nReadDataErrors;
  OTX_UINT32  nReadDataLoops;
  OTX_UINT32  nNoEventCount;
  OTX_UINT32  nSavedFifoPeakBytes;
  OTX_UINT32  nSavedFifoPeakFrames;
} sBurstStatT;
static sBurstStatT g_sBurstStat;

FILE       *g_pRxFile = NULL;
//FILE       *g_pRxFileLi0 = NULL;
//FILE       *g_pRxFileLi1 = NULL;
//FILE       *g_pRxFileLi2 = NULL;
//FILE       *g_pRxFileLi3 = NULL;
FILE *f;
OTX_INT32  g_nTsToRecord = 10; 

static OTX_UINT32 g_cbBurstBufSize = 0;  // Set to a valid value later

static OTX_UINT8 *g_sqBurstDmaBuf1 = NULL;
static OTX_UINT8 *g_sqBurstDmaBuf2 = NULL;

OTX_UINT8 *g_pCurrentTxBuf = g_sqBurstDmaBuf1;
OTX_UINT8 *g_pCurrentRxBuf = g_sqBurstDmaBuf2;

OTX_UINT8 g_sqSingleTsBuf[DEMO_BURST_BUF_SIZE/128];

OTX_BOOL    g_bLoopTransceiverMode = OTX_FALSE;
OTX_BOOL    g_bFileTransceiverMode = OTX_FALSE;


struct {
  OTX_HANDLE hBoard;
  OTX_HANDLE hAspBurst;
  OTX_HANDLE hQuadE1T1;
  OTX_HANDLE hLi[8];
} g_hPhysDevs;

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;

OTX_UINT32 g_coFramesInPacket = 0;
OTX_UINT8  g_sqPacketBuf[DEMO_MAX_FRAMES_IN_PACKET * 128];

#ifdef DEMO_UDP_PACKETS
otx::udp_socket g_hSocket;
sockaddr_in     g_addrDest;
#endif

/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;
                               
  // Create a notification Event for low priority events
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    // D16V86 apps cannot use event objects
    g_hNotificationEvent = (OTX_EVENT) NULL;      
    result = OTX_S_OK;
  }

  DemoCheckRc(result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); 
  DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); 
  DemoCheckRc(result);

    // Create a notification Event high priority (Burst) events
  result = OtxDrvCreateEvent(&g_hBurstNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    // D16V86 apps cannot use event objects
    g_hBurstNotificationEvent = (OTX_EVENT) NULL;      
    result = OTX_S_OK;
  }

  DemoCheckRc(result);

  // Create Event Queue for burst device notifications
  result = OtxDrvCreateEventQueue(128, g_hBurstNotificationEvent, &g_hBurstEventQueue); 
  DemoCheckRc(result);

  // Enable Event Queue for the burst device
  result = OtxDrvEnable(g_hBurstEventQueue, OTX_TASK_SYNC,0); 
  DemoCheckRc(result);

}



/*******************************************************************************

  @func
    Open the Physical Devices within Alvis CSI

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_UINT32 nDeviceId;
  OTX_INT32 i;
  OTX_UINT32 cbBurstBufSize = DEMO_BURST_BUF_SIZE;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
    DemoLogf("Opening Alvis-2-PBX board (test configuration)\n");
    g_demoEnv.coBurstHwyTimeslots = 128;
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
    DemoLogf("Opening Alvis-2-CSI board\n");
    g_demoEnv.coBurstHwyTimeslots = 128;
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
    DemoLogf("Opening Alvis-4-CSI board\n");
    g_demoEnv.coBurstHwyTimeslots = 128;
    g_demoEnv.coLi = 4;
    break;
  case OTX_DEVICE_DMP_ALVIS_4_PBX:
    DemoLogf("Opening Alvis-4-PBX board (test configuration)\n");
    g_demoEnv.coBurstHwyTimeslots = 128;
    g_demoEnv.coLi = 4;
    break;
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
    DemoLogf("Opening Alvis-4M-CSI board\n");
    g_demoEnv.coBurstHwyTimeslots = 128;
    g_demoEnv.coLi = 4;
    break;
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
    DemoLogf("Opening Alvis-8-CSI board\n");
    cbBurstBufSize = DEMO_BURST_BUF_SIZE * 2;
    g_demoEnv.coBurstHwyTimeslots = 256;
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
    DemoLogf("Opening Alvis-8-PBX board (test configuration)\n");
    cbBurstBufSize = DEMO_BURST_BUF_SIZE * 2;
    g_demoEnv.coBurstHwyTimeslots = 256;
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
    DemoLogf("Opening Alvis-8M-CSI board\n");
    cbBurstBufSize = DEMO_BURST_BUF_SIZE * 2;
    g_demoEnv.coBurstHwyTimeslots = 256;
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_DMP_ALVIS:
    DemoLogf("Opening Alvis-DMP (hosted by ASM or PCIe)\n");
    g_demoEnv.coBurstHwyTimeslots = 128;
    g_demoEnv.coLi = 0;
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    DemoExitQuestion();
    break;
  }

  // Open the DaVinci Nic and the DMP device (hosted by the NIC)
#if 1
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, 0, 0, g_hEventQueue, &g_hPhysDevs.hBoard); DemoCheckRc(result);
#else
  result = OtxDrvGetDeviceHandle(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, 0, &g_hPhysDevs.hBoard); DemoCheckRc(result);
#endif

  // Open the ASP Burst device
  DemoLogf("Opening Burst device\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoard, OTX_DEVICE_BURST, 0, 0, g_hBurstEventQueue, &g_hPhysDevs.hAspBurst); DemoCheckRc(result);

  if (g_demoEnv.bDisableBurstRxFifo) {
    DemoLogf("Disabling RX FIFO\n"); // Disable RX FIFO by setting it to 0
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_RECEIVER_FIFO_SIZE, 0); DemoCheckRc(result);
  }


  {
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_TX_BUFFER_SIZE, &g_cbBurstBufSize); DemoCheckRc(result);
    DemoLogf("Default Burst Buffer Size       = %d\n", g_cbBurstBufSize);

    g_cbBurstBufSize = cbBurstBufSize;
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_TX_BUFFER_SIZE, g_cbBurstBufSize); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_TX_BUFFER_SIZE, &g_cbBurstBufSize); DemoCheckRc(result);
    DemoLogf("Changed Burst Buffer Size       = %d\n", g_cbBurstBufSize);

    g_sqBurstDmaBuf1 = (OTX_UINT8 *) malloc(g_cbBurstBufSize);
    g_sqBurstDmaBuf2 = (OTX_UINT8 *) malloc(g_cbBurstBufSize);

#if 1 //
    DemoLogf("*********************** Two dummy mallocs\n");
    malloc(g_cbBurstBufSize);
    malloc(g_cbBurstBufSize);
#endif

    g_pCurrentTxBuf = g_sqBurstDmaBuf1;
    g_pCurrentRxBuf = g_sqBurstDmaBuf2;

    if (!g_sqBurstDmaBuf1 || !g_sqBurstDmaBuf2) {
      DemoLogf("ERROR: Failed to allocate burst buffers\n", g_demoEnv.nBoardType);
      DemoExitQuestion();
    }
  }

  {
    OTX_UINT32 nFifoSize;
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_RECEIVER_FIFO_SIZE, &nFifoSize); DemoCheckRc(result);
    DemoLogf("FIFO Size (bytes)     = %d\n", nFifoSize);
  }


  // Optionally (mainly for testing, or for improving system performance for single direction burst applications,
  // disable one direction of the burst device
  if (g_demoEnv.bDisableBurstRx) {
    DemoLogf("Burst RX interrupts will be turned off when the Burst device is enabled\n");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_RECV_DISABLED,  OTX_TRUE); DemoCheckRc(result);
  }
  if (g_demoEnv.bDisableBurstTx) {
    DemoLogf("Burst TX interrupts will be turned off when the Burst device is enabled\n");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_SEND_DISABLED,  OTX_TRUE); DemoCheckRc(result);
  }


  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoard, OTX_ATTR_BRD_DEVICE_ID, &nDeviceId); DemoCheckRc(result);

  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoard, OTX_ATTR_BRD_SUBSYSTEM_ID,  &g_nSubsystemId); DemoCheckRc(result);
  DemoLogf("Board Device ID: 0x%.4X, SubsystemId: 0x%.4X\n", nDeviceId, g_nSubsystemId);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_DMP_ALVIS_4_PBX:
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
    DemoLogf("Open OTX_DEVICE_QUAD_T1E1\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoard, OTX_DEVICE_QUAD_T1E1, 0, 0, g_hEventQueue, &g_hPhysDevs.hQuadE1T1); DemoCheckRc(result);

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1)\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
    DemoLogf("Open OTX_DEVICE_DUAL_T1E1\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoard, OTX_DEVICE_DUAL_T1E1, 0, 0, g_hEventQueue, &g_hPhysDevs.hQuadE1T1); DemoCheckRc(result);

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_DUAL_T1E1)\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  }

}


/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);
  
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(result);

  }

}

/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = g_demoEnv.nLineLength;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;    // CCS, CAS_BR;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_TRUE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_TRUE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = g_demoEnv.nLineLength;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}



/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Enable Board
  result = OtxDrvEnable(g_hPhysDevs.hBoard,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
  case OTX_DEVICE_DMP_ALVIS_4_PBX:
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
    // Enable Quad E1T1 device
    DemoLogf("Enable T1E1 host device\n");
    result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Enable Li%d\n", i);
      result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
    break;
  default:
    break;
  }


}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 
  // (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a 
  // very accurate way to report timestamp)
  nTimeStampMs.nUint64 = (nTimeStamp.nUint64 / 10000) - 
                         (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}



/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;

  ui32 = g_cbBurstBufSize;
  
  otxRc = OtxBurstSendData(g_hPhysDevs.hAspBurst, nSourceId, g_pCurrentTxBuf, &ui32); 
  
  if (OTX_SUCCESS(otxRc)) {
    if(ui32 != g_cbBurstBufSize) {
      DemoLogf("DemoBurstSender[%d]: OtxBurstSendData did not copy entire buffer (%d).\n", nSourceId, ui32);
    }
  } else {
    g_sBurstStat.nWriteDataErrors++;
  }

  if (g_bLoopTransceiverMode) {
    // Point to the other buffer
    if (g_pCurrentTxBuf == g_sqBurstDmaBuf1) {
      g_pCurrentTxBuf = g_sqBurstDmaBuf2;
    } else {
      g_pCurrentTxBuf = g_sqBurstDmaBuf1;
    }
  }


}


/*******************************************************************************
  @func
    Burst Receiver Function

    Reads a block of data from the Burst device. The block is formatted in the
    following way:

    Alvis-4-CSI, Alvis-4M-CSI (highway speed is 8Mbit/s, 128 timeslots)

      block index mod 128:    0   1   2   3    4   5   6   7  ... 124  125  126  127
      Timeslot(0-127):       TS0 TS0 TS0 TS0  TS1 TS1 TS1 TS1 ... TS31 TS31 TS31 TS31
      Li(0-3):               Li3 Li2 Li1 Li0  Li3 Li2 Li1 Li0 ... Li3  Li2  Li1  Li0

    Alvis-8-CSI, Alvis-8M-CSI (highway speed is 16Mbit/s, 256 timeslots)

      block index mod 256:    0   1   2   3   4   5   6   7    8   9   10  11  12  13  15  16 ... 248  249  250  251  252  253  254  255
      Timeslot(0-255):       TS0 TS0 TS0 TS0 TS0 TS0 TS0 TS0  TS1 TS1 TS1 TS1 TS1 TS1 TS1 TS1 ... TS31 TS31 TS31 TS31 TS31 TS31 TS31 TS31
      Li(0-7):               Li3 Li2 Li1 Li0 Li7 Li6 Li5 Li4  Li3 Li2 Li1 Li0 Li7 Li6 Li5 Li4 ... Li3  Li2  Li1  Li0  Li7  Li6  Li5  Li4

*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;
  int cbWritten;

  ui32 = g_cbBurstBufSize;
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hAspBurst, nSourceId, g_pCurrentRxBuf, &ui32, &nFrameRefId, &dtFrameRefTime); 
  
  while (otxRc == OTX_S_OK) {
    if (ui32 != g_cbBurstBufSize) {
      DemoLogf("DemoBurstReceiver[%d]: sizeof(g_sqBurstDmaBuf2) [%d] != ui32 [%d].\n", nSourceId, g_cbBurstBufSize, ui32);
    }
    
    if (g_pRxFile && g_bFileTransceiverMode) {
#if 0
      cbWritten = fwrite(g_pCurrentRxBuf, 1, g_cbBurstBufSize, g_pRxFile);
      if (cbWritten != (int)g_cbBurstBufSize) {
        DemoLogf("DemoBurstReceiver[%d]: Could not write the expected number of bytes to the file (%d).\n", nSourceId, ui32);
      }
#else
      int byteCount = 0;
      int x;
        
      byteCount = 0;
      for (x = 7; x < (int)ui32; x += g_demoEnv.coBurstHwyTimeslots) {
        g_sqSingleTsBuf[byteCount++] = g_pCurrentRxBuf[x];
      }

      cbWritten = fwrite(g_sqSingleTsBuf, sizeof(unsigned char), ui32/g_demoEnv.coBurstHwyTimeslots, g_pRxFile);
      if (cbWritten != byteCount) {
        DemoLogf("Error: Could not write data to file.  (ui32/%d=%d, cbWritten=%d) \n", g_demoEnv.coBurstHwyTimeslots, ui32/g_demoEnv.coBurstHwyTimeslots, cbWritten);
      }
#endif
      //for (i = 0; i < (int)g_cbBurstBufSize; i += 128) {
      //  fwrite(&g_pCurrentRxBuf[i + 3 + nTs*4], 1, 1, g_pRxFileLi0);
      //  fwrite(&g_pCurrentRxBuf[i + 2 + nTs*4], 1, 1, g_pRxFileLi1);
      //  fwrite(&g_pCurrentRxBuf[i + 1 + nTs*4], 1, 1, g_pRxFileLi2);
      //  fwrite(&g_pCurrentRxBuf[i + 0 + nTs*4], 1, 1, g_pRxFileLi3);
      //}
    }
    
#ifdef DEMO_UDP_PACKETS
    {
      
      static unsigned char Seq=0;
      static int cntSeq = 0;
      int ret;
      
#if 0
      {
        int j;
        for(int i=3,j=0;i<g_cbBurstBufSize;i+=4,j++) {//128) {
          
          if(j%32==0) {
            continue;
          }
          if(Seq!=g_pCurrentRxBuf[i]) {
            if(!((cntSeq++) %10)) DemoLogf("Seq! %d\n",cntSeq);
            Seq=g_pCurrentRxBuf[i]+1;
          }else{
            Seq++;
          }
          
        }
      }
#endif
      //if(cntSeq++ % 10 == 0)  {
        g_pCurrentRxBuf[0] = Seq++;
      
        ret = g_hSocket.send_to((OTX_CHAR*)g_pCurrentRxBuf, g_cbBurstBufSize, (sockaddr*)&g_addrDest, sizeof(g_addrDest));
        if(ret != g_cbBurstBufSize) {
          DemoLogf("send_to return %d\n", ret);
        }
      //}
      //    else DemoLogf("OK\n");
    }
#endif // DEMO_UDP_PACKETS

    ui32 = g_cbBurstBufSize;
    otxRc = OtxBurstReadDataEx(g_hPhysDevs.hAspBurst, nSourceId, g_pCurrentRxBuf, &ui32, &nFrameRefId, &dtFrameRefTime); 
    if (otxRc == OTX_S_OK) {
      g_sBurstStat.nReadDataLoops++;
    }

  } // while

  if (!OTX_SUCCESS(otxRc)) {
    g_sBurstStat.nReadDataErrors++;
  }

  if (g_bLoopTransceiverMode) {
    // Point to the other buffer
    if (g_pCurrentRxBuf == g_sqBurstDmaBuf1) {
      g_pCurrentRxBuf = g_sqBurstDmaBuf2;
    } else {
      g_pCurrentRxBuf = g_sqBurstDmaBuf1;
    }
  }

}



#define DEMO_EVENT_DATA_BUF_SIZE 10


/*******************************************************************************
  @func
    Idle Function that check for Events from the Burst device only.
    This thread runs at high priority, since the burst device has to be serviced within
    8 ms (assuming 8k byte ping/pong buffers).
*******************************************************************************/
void  *DemoBurstIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;


// Increase the priority
  struct sched_param param;  // scheduling priority
  int policy;     // scheduling policy

  // Get the current thread id
  pthread_t thread_id = pthread_self();

  //DemoLogf("Get scheduling parameters\n");
  pthread_getschedparam(thread_id, &policy, &param);

  DemoLogf("The thread scheduling parameters indicate priority = %d\n", param.sched_priority);

  // To set the scheduling priority of the thread
  policy = SCHED_RR;     // scheduling policy
  param.sched_priority = 90;
  pthread_setschedparam(thread_id, policy, &param);

  while (1) {
    OtxDrvWaitForSingleEvent(g_hBurstNotificationEvent, OTX_TIME_INFINITY);

    do {

      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      result = OtxDrvGetEventDataEx(g_hBurstEventQueue, 
                                    sqEventData, 
                                    DEMO_EVENT_DATA_BUF_SIZE, 
                                    &coEventsFetched); 
      DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)

        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          if (sqEventData[i].m_eDeviceType == OTX_DEVICE_BURST) {
            // Handle the events from the Burst device here (since they are frequent and need special attention)
            switch (sqEventData[i].m_nCode) {
            case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
              g_sBurstStat.nRxDataAvailable++;
              DemoBurstReceiver(0);
              break;
            case OTX_BURST_SENDER_EC_BUF_DEPLETED:
              g_sBurstStat.nTxBufDepleted++;
              DemoBurstSender(0);
              break;
            case OTX_BURST_EC_BUFFER_OVERRUN:  // Receiver data overrun error - data lost
              g_sBurstStat.nSenderOverRunErrors++;
              DemoLogf("ASP Burst Event: Receiver OTX_BURST_EC_BUFFER_OVERRUN.\n");
              DemoBurstReceiver(0); // Clear the receiver overrun condition so that we can receive this event again, if we overrun again
              break;
            case OTX_BURST_EC_BUFFER_UNDERRUN: // Sender data underrun error
              g_sBurstStat.nReceiverUnderRunErrors++;
              DemoLogf("ASP Burst Event: Sender OTX_BURST_EC_BUFFER_UNDERRUN\n");
              DemoBurstSender(0); // Clear the sender underrun condition so that we can receive this event again, if we underrun again
              break;
            default:
              DemoLogf("DemoBurstIdleFunction eDeviceType 0x%X nSourceId 0x%X nCode 0x%X\n", 
                sqEventData[i].m_eDeviceType,
                sqEventData[i].m_nSourceId,
                sqEventData[i].m_nCode);
              break;
            }
          } else {
            DemoLogf("DemoBurstIdleFunction eDeviceType 0x%X nSourceId 0x%X nCode 0x%X\n", 
              sqEventData[i].m_eDeviceType,
              sqEventData[i].m_nSourceId,
              sqEventData[i].m_nCode);
          }
        }

      } else {
        g_sBurstStat.nNoEventCount++;
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;

}



/*******************************************************************************
  @func
    
*******************************************************************************/
void  *DemoEatMemory(
  void *lpParameter   // thread data
)
{

  while (1) {
    malloc(32768);
    OsaThreadSleep(10);
  }

  return 0;

}





/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");

    do {

      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      result = OtxDrvGetEventDataEx(g_hEventQueue, 
                                    sqEventData, 
                                    DEMO_EVENT_DATA_BUF_SIZE, 
                                    &coEventsFetched); 
      DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)

        for (i = 0; i < coEventsFetched; i++) {
          // Act on the event data.
          DemoPrintEvent(&sqEventData[i]);
        }

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event. 
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_DMP_ALVIS:
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
  case OTX_DEVICE_DMP_ALVIS_4_PBX:
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
    {
      switch (pEventData->m_nCode) {
      case OTX_BRD_EC_CORE2CORE_INTERRUPT:
        {
          DemoLogf("Core-to-Core interrupt\n");  // OTX Internal use
        }
        break;
      default:
        DemoLogf("DemoPrintEvent eDeviceType 0x%X nSourceId 0x%X nCode 0x%X\n", 
          pEventData->m_eDeviceType,
          pEventData->m_nSourceId,
          pEventData->m_nCode);
        break;
      }
    }
    break;

  case OTX_DEVICE_LI_T1E1:
    // Event from Li
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (!strncmp(msgStr, "Invalid", strlen("Invalid"))) {
      DemoLogf("E1/T1[%d] Code is %d\n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    DemoLogf("DemoPrintEvent eDeviceType 0x%X nSourceId 0x%X nCode 0x%X\n", 
      pEventData->m_eDeviceType,
      pEventData->m_nSourceId,
      pEventData->m_nCode);
    break;
  }


}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr         = NULL;
  envP->bExitOnError        = OTX_FALSE;
  envP->nBoardType          = OTX_DEVICE_UNKNOWN;
  envP->coLi                = 4;
  envP->coBurstHwyTimeslots = 128;
  envP->eT1E1Mode           = OTX_T1E1_MODE_T1;
  envP->nLineLength         = 10;
  envP->bReceiveEqualizer   = OTX_FALSE;
  envP->eT1LineCode         = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat      = OTX_T1_FF_ESF;
  envP->eE1LineCode         = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat      = OTX_E1_FF_CRC4_MULTIFRAME;
  envP->bDisableBurstTx     = OTX_FALSE;
  envP->bDisableBurstRx     = OTX_FALSE;
  envP->bDisableBurstRxFifo = OTX_FALSE;

  strcpy(envP->rxFileName, DEMO_ASP_BURST_RX_FILE_NAME);

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      //OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'R') {   // Rx data file
          strcpy(envP->rxFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'D') {   // Disable one direction of burst
          if (*(argv[i]+2) == 'T') {   // TX
            envP->bDisableBurstTx = OTX_TRUE;
          } else if (*(argv[i]+2) == 'R') {   // T1
            envP->bDisableBurstRx = OTX_TRUE;
          } else {
            DemoLogf("Error: Invalid Burst direction-disabling command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'X') {   // Disable RX Fifo
          envP->bDisableBurstRxFifo = OTX_TRUE;
        } else if (*(argv[i]+1) == 'T') {   // Board Type

          if (*(argv[i] + 2) == 'D') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4_CSI;
          } else if (*(argv[i] + 2) == 'K') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4M_CSI;
          } else if (*(argv[i] + 2) == 'X') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_8_CSI;
          } else if (*(argv[i] + 2) == 'Q') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_8M_CSI;
          } else if (*(argv[i] + 2) == 'B') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_2_CSI;
          } else if (*(argv[i] + 2) == 'G') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_2_PBX;
          } else if (*(argv[i] + 2) == 'I') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4_PBX;
          } else if (*(argv[i] + 2) == 'J') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_8_PBX;
          } else if (*(argv[i] + 2) == 'A') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /T[D|K|X|Q|B|A]      : Board Type used for the demo, where\n");
  DemoLogf("                    D  = Use Alvis-4-CSI board\n");
  DemoLogf("                    K  = Use Alvis-4M-CSI board\n");
  DemoLogf("                    X  = Use Alvis-8-CSI board\n");
  DemoLogf("                    Q  = Use Alvis-8M-CSI board\n");
  DemoLogf("                    B  = Use Alvis-2-CSI board\n");
  DemoLogf("                    G  = Use Alvis-2-PBX Board (test config)\n");
  DemoLogf("                    I  = Use Alvis-4-PBX Board (test config)\n");
  DemoLogf("                    J  = Use Alvis-8-PBX Board (test config)\n");
  DemoLogf("                    A  = Use Alvis-DMP (hosted on ASM or PCIe)\n");
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /R<fileName>   : Rx Burst Data file name (default is %s)\n", DEMO_ASP_BURST_RX_FILE_NAME);
  DemoLogf(" /X             : Disable RX FIFO (lower CPI usage, but underruns are possible)\n");
  DemoLogf(" /D[T|R]        : Disable one direction of the Burst device\n");
  DemoLogf("                  T = Disable Burst TX\n");
  DemoLogf("                  R = Disable Burst RX\n");
  DemoLogf(" /ME            : E1 Li mode\n");
  DemoLogf(" /MT            : T1 Li mode\n");
  DemoLogf(" /H             : Longhaul mode (default is shorthaul mode)\n");
  DemoLogf(" /F[E|S|9]      : T1 Frame Format, where\n");
  DemoLogf("                    E = ESF (default)\n");
  DemoLogf("                    4 = F4\n");
  DemoLogf("                    S = SF [D3/D4] (F12)\n");
  DemoLogf("                    9 = SLC96 (F72)\n");
  DemoLogf(" /J[A|B]        : T1 Line Code, where\n");
  DemoLogf("                    A = AMI\n");
  DemoLogf("                    B = B8ZS (default)\n");
  DemoLogf(" /G[D|C|G]      : E1 Frame Format, where\n");
  DemoLogf("                    D = Doubleframe\n");
  DemoLogf("                    C = CRC4 Multiframe (default)\n");
  DemoLogf("                    G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]        : E1 Line Code, where\n");
  DemoLogf("                    A = AMI\n");
  DemoLogf("                    H = HDB3 (default)\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");

  DemoLogf("\nPress Any Key to Exit...\n"); OsaConsoleGetCh();

}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile, formatStr, argList);

  va_end(argList);

}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(0, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", 
      OTX_RESULT_CODE(result), 
      OTX_RESULT_SOURCE_ID(result), 
      OTX_RESULT_LINE_NUM(result), 
      msgStr);

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" e/D Enable(e)/Diable(E) the ASP DMA Burst Stats\n");
    DemoLogf(" b/B Start(b)/Stop(B) sending/receiving files \n");
    DemoLogf(" x/X Start(x)/Stop(X) ASP DMA Loop mode (RX->TX)\n");
    DemoLogf(" s   Print ASP DMA Burst Stats\n");
    DemoLogf(" l   Print E1/T1 Line status\n");
    DemoLogf(" n   Choose timeslot[%d] for recording (/tmp/rxLiXOneTs.raw files)\n", g_nTsToRecord);
    DemoLogf(" R   Reset the Asp Burst device (FIFO and counters)\n");
    DemoLogf(" d   Send DMP Debug Command\n");
    DemoLogf(" D   Send ASP Debug Command\n");
    DemoLogf(" P   Start eating memory\n");
    DemoLogf(" c   Set Clock Source\n");
    break;

  case 'P':
    {
      DemoLogf("Starting threads that eats memory (does not free it)\n");
      OsaThreadStart(DemoEatMemory);
    }
    break;


  case 'd':
    {
      OTX_UINT32 nCmd;

      DemoLogf("DMP Debug Command. Select Command:\n");
      DemoLogf("0 = Read GPIO registers\n");
      DemoLogf("1 = Issue Core interrupt\n");

      nCmd = OsaConsoleGetCh() - '0';

      DemoLogf("Command value 0x%X sent\n", nCmd);

      OtxDrvDbgDump(g_hPhysDevs.hBoard, nCmd);

      DemoLogf("Done.\n");
    }
    break;


  case 'D':
    {
      OTX_UINT32 nCmd = 0xFFFF0000; // Do nothing
      char strBuf[32];

      DemoLogf("Select Command:\n");
      DemoLogf("100 = Display McBSP Registers\n");
      DemoLogf("101 = Generate an OTX_BURST_EC_BUFFER_UNDERRUN event\n");
      DemoLogf("1 = Display ASP status counters\n");
      DemoLogf("2 = Fill ASP Transmit buffer with 0xA5 value\n");
      DemoLogf("3 = Fill ASP Transmit buffer with an 8 bit count\n");
      DemoLogf("4 = Print 128 bytes from RxBuf\n");
      DemoLogf("5 = IPR clearing\n");
      DemoLogf("6 = Set one bit in EER\n");

      DemoLogf("Enter command: ");

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nCmd = strtoul(strBuf, NULL, 0);
      }

      if(nCmd != 0xFFFF0000) {
        DemoLogf("Command value 0x%X sent\n", nCmd);

        OtxDrvDbgDump(g_hPhysDevs.hAspBurst, nCmd);
      }

      DemoLogf("Done.\n");
    }
    break;



  case 'n':
    {
      char strBuf[12];
      OTX_INT32 nTs;
      DemoLogf("Choose TS [0-31] for recording (/tmp/rxLiXOneTs.raw files) [%d]\n", g_nTsToRecord);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTs = (OTX_UINT8)strtoul(strBuf, NULL, 0);

        if (nTs >= 0 && nTs <= 31) {
          g_nTsToRecord = nTs;
          DemoLogf("TS%d will now be recorded when 'b' is pressed\n", g_nTsToRecord);
        } else {
          DemoLogf("Invalid TS [%d]. Must be in the range of 0-31\n", nTs);
        }
      } else {
        DemoLogf("No change\n");
      }

    }
    break;

  case 'N':
    {
      FILE *pFile;
      pFile = fopen("/tmp/dump", "wb");
      if(pFile) {
        fwrite(g_pCurrentRxBuf, g_cbBurstBufSize, 1, pFile);
	fclose(pFile);
      }
    }
    break;

  case 'e':
    {
      OtxDeviceStateE  eDevState;
      //int x;
      //int nLi;
      //int nTs;
      //OTX_UINT8 nByteVal = 0;

      result = OtxDrvGetState(g_hPhysDevs.hAspBurst, &eDevState); DemoCheckRc(result);
      
      g_bLoopTransceiverMode = OTX_FALSE;
      g_bFileTransceiverMode = OTX_FALSE;

#if 1
      memset(g_pCurrentTxBuf, 0xA5, g_cbBurstBufSize);
#else
      // Send  the highway TS number in the timeslot
      for (x = 0; x < g_cbBurstBufSize; /**/) {
        for (nTs = 0; nTs < 128; nTs++) {
          g_pCurrentTxBuf[x++] = nTs;
        }
      }
#endif


      if (eDevState == OTX_DEVICE_WORKING) {
        DemoLogf("The ASP DMA Burst Device is already enabled. No action\n");
        break;
      }

      DemoLogf("Enabling ASP DMA Burst Device. g_cbBurstBufSize=%d\n", g_cbBurstBufSize);
      result = OtxDrvEnable(g_hPhysDevs.hAspBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
    break;

  case 'X':
  case 'B':
  case 'E':
    {
      OtxDeviceStateE  eDevState;

      result = OtxDrvGetState(g_hPhysDevs.hAspBurst, &eDevState); DemoCheckRc(result);
      
      if (eDevState != OTX_DEVICE_WORKING) {
        DemoLogf("The ASP DMA Burst Device is not enabled. Can't disable it. No action\n");
        break;
      }

      g_bLoopTransceiverMode = OTX_FALSE;
      g_bFileTransceiverMode = OTX_FALSE;
      result = OtxDrvDisable(g_hPhysDevs.hAspBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

      if (g_pRxFile) {
        OsaThreadSleep(300);  // Give the burst device some time to complete any pending events (if any)
        DemoLogf("Closing %s file\n", g_demoEnv.rxFileName);
        fclose(g_pRxFile);
        //fclose(g_pRxFileLi0);
        //fclose(g_pRxFileLi1);
        //fclose(g_pRxFileLi2);
        //fclose(g_pRxFileLi3);
        g_pRxFile = NULL;
      }

      DemoLogf("Disabled ASP DMA Burst Device.\n");
      DemoLogf("Done.\n");
    }
    break;

  case 'b':
    {
      int x;
      OtxDeviceStateE  eDevState;

      DemoLogf("Opening %s file for writing (g_cbBurstBufSize=%d)\n", g_demoEnv.rxFileName, g_cbBurstBufSize);
      g_pRxFile = fopen(g_demoEnv.rxFileName, "wb");
      if (!g_pRxFile) {
        DemoLogf("ERROR: Failed to open %s file.\n", g_demoEnv.rxFileName);
        break;
      }

      //DemoLogf("Opening Li (%s) files for writing (timeslot=%d)\n", "/tmp/rxLiXOneTs.raw", g_nTsToRecord);
      //g_pRxFileLi0 = fopen("/tmp/rxLi0OneTs.raw", "wb");
      //g_pRxFileLi1 = fopen("/tmp/rxLi1OneTs.raw", "wb");
      //g_pRxFileLi2 = fopen("/tmp/rxLi2OneTs.raw", "wb");
      //g_pRxFileLi3 = fopen("/tmp/rxLi3OneTs.raw", "wb");

      g_pCurrentTxBuf = g_sqBurstDmaBuf1;
      g_pCurrentRxBuf = g_sqBurstDmaBuf2;

      // Send a gamma pattern, which means that we increase the value by one for each frame (continuously until aborted)
      for (x = 0; x < (int)g_cbBurstBufSize; x++) {
        g_pCurrentTxBuf[x] = (OTX_UINT8) x;
      }
      g_bFileTransceiverMode = OTX_TRUE;

      result = OtxDrvGetState(g_hPhysDevs.hAspBurst, &eDevState); DemoCheckRc(result);
      
      if (eDevState != OTX_DEVICE_WORKING) {
        DemoLogf("Enabling ASP DMA Burst Device\n");
        result = OtxDrvEnable(g_hPhysDevs.hAspBurst, OTX_TASK_SYNC, 5000L);  DemoCheckRc(result);
      }
    }
    break;

  case 'x':
    {
      OtxDeviceStateE  eDevState;

      DemoLogf("Burst loopback mode (g_cbBurstBufSize=%d)\n", g_cbBurstBufSize);

      g_pCurrentTxBuf = g_sqBurstDmaBuf1;
      g_pCurrentRxBuf = g_sqBurstDmaBuf2;

      g_bLoopTransceiverMode = OTX_TRUE;

      result = OtxDrvGetState(g_hPhysDevs.hAspBurst, &eDevState); DemoCheckRc(result);
      
      if (eDevState != OTX_DEVICE_WORKING) {
        DemoLogf("Enabling ASP DMA Burst Device\n");
        result = OtxDrvEnable(g_hPhysDevs.hAspBurst, OTX_TASK_SYNC, 5000L);  DemoCheckRc(result);
      }
    }
    break;


  case 'R':
    {
      DemoLogf("Reseting FIFO and counters in ASP DMA Burst Device\n");
      memset(&g_sBurstStat, 0, sizeof(g_sBurstStat));
      result = OtxDrvReset(g_hPhysDevs.hAspBurst);  DemoCheckRc(result);
    }
    break;




  case 's':
    {
      OTX_UINT32 nVal;

      DemoLogf("ASP DMA Burst Stats\n");
      DemoLogf("TX:\n");

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_STARTS, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Start             = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_COMPLETES, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Complete          = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_PARTIAL, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Ping Buf Ready    = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_FULL, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Pong Buf Ready    = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_AMOUNT, &nVal); DemoCheckRc(result);
      DemoLogf("  Byte Count            = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_ERRORS, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Error             = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_ABORTS, &nVal); DemoCheckRc(result);
      DemoLogf("  Sync Errors           = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS, &nVal); DemoCheckRc(result);
      DemoLogf("  HW Underruns          = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS_USER, &nVal); DemoCheckRc(result);
      DemoLogf("  User Underruns        = %d\n", nVal);
      //result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_SEND_RESERVED0, &nVal); DemoCheckRc(result);
      //DemoLogf("  Reserved0             = %d\n", nVal);

      DemoLogf("RX:\n");

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_STARTS, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Start             = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Complete          = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_PARTIAL, &nVal); DemoCheckRc(result);  // Number of Ping buf ready
      DemoLogf("  DMA Ping Buf Ready    = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_FULL, &nVal); DemoCheckRc(result);     // Number of Pong buf ready
      DemoLogf("  DMA Pong Buf Ready    = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_AMOUNT, &nVal); DemoCheckRc(result);
      DemoLogf("  Byte Count            = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_ERRORS, &nVal); DemoCheckRc(result);
      DemoLogf("  DMA Error             = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_ABORTS, &nVal); DemoCheckRc(result);
      DemoLogf("  Sync Errors           = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS, &nVal); DemoCheckRc(result);
      DemoLogf("  HW Overruns           = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS_USER, &nVal); DemoCheckRc(result);
      DemoLogf("  User Overruns         = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY_PARTIAL, &nVal); DemoCheckRc(result);
      DemoLogf("  FIFO Overruns         = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_RECEIVER_FIFO_SIZE, &nVal); DemoCheckRc(result);
      DemoLogf("  FIFO Size (bytes)     = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_RECEIVER_FIFO_COUNT, &nVal); DemoCheckRc(result);
      DemoLogf("  Current bytes in FIFO = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_RECEIVER_FIFO_PEAK, &nVal); DemoCheckRc(result);
      DemoLogf("  Peak bytes in FIFO    = %d (since last 's' command)\n", nVal);
      if (nVal > g_sBurstStat.nSavedFifoPeakBytes) {
        g_sBurstStat.nSavedFifoPeakBytes = nVal;
      }
      DemoLogf("  Peak bytes in FIFO    = %d (since last 'R' [Reset] command)\n", g_sBurstStat.nSavedFifoPeakBytes);

#if 1 // Just for debugging. The meaning of RESERVED attributes can changed from one driver release to the next
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_RESERVED1, &nVal); DemoCheckRc(result);
      DemoLogf("  Peak frames in FIFO   = %d (since last 's' command)\n", nVal);
      if (nVal > g_sBurstStat.nSavedFifoPeakFrames) {
        g_sBurstStat.nSavedFifoPeakFrames = nVal;
      }
      DemoLogf("  Peak frames in FIFO   = %d (since last 'R' [Reset] command)\n", g_sBurstStat.nSavedFifoPeakFrames);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_RESERVED0, &nVal); DemoCheckRc(result);
      DemoLogf("  Reserved0             = %d\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_RESERVED6, &nVal); DemoCheckRc(result);
      DemoLogf("  Reserved6             = %d (FIFO unexpectedly full)\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_RESERVED7, &nVal); DemoCheckRc(result);
      DemoLogf("  Reserved7             = %d (Extra frames inserted)\n", nVal);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hAspBurst, OTX_ATTR_BURST_STAT_RECV_RESERVED5, &nVal); DemoCheckRc(result);
      DemoLogf("  Reserved5             = %d (Extra frame treatment)\n", nVal);
#endif

      DemoLogf("Event Stats\n");
      DemoLogf("  Rx Data Available     = %d\n", g_sBurstStat.nRxDataAvailable);
      DemoLogf("  Tx Buf Depleted       = %d\n", g_sBurstStat.nTxBufDepleted);
      DemoLogf("  Rx Overruns           = %d\n", g_sBurstStat.nSenderOverRunErrors);
      DemoLogf("  Tx Underruns          = %d\n", g_sBurstStat.nReceiverUnderRunErrors);
      DemoLogf("  NoEvent Count         = %d\n", g_sBurstStat.nNoEventCount);
      
      DemoLogf("Other Counters\n");
      DemoLogf("  Write Data Errors     = %d\n", g_sBurstStat.nWriteDataErrors);
      DemoLogf("  Read Data Errors      = %d\n", g_sBurstStat.nReadDataErrors);
      DemoLogf("  Read Data Loops       = %d\n", g_sBurstStat.nReadDataLoops);

    }
    break;


  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;
      OTX_INT32 i;

      if (g_demoEnv.nBoardType != OTX_DEVICE_DMP_ALVIS) {
        for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
          DemoLogf("Line status of Li #%d is ", nLi);
          result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(result);
          DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
        }
      }

      DemoLogf("Frame snapshot of ASP DMA Rx Buf:\n");
      for (i = 0; i < g_demoEnv.coBurstHwyTimeslots; i++) {
        if ((i%8) == 0) DemoLogf("[%.3d]: ", i);
        DemoLogf("%.2X ", g_pCurrentRxBuf[i]);
        if (((i+1)%8) == 0) DemoLogf("\n");
      }
      DemoLogf("\n");
            
    }
    break;


  case 'c':
    {
      OtxBrdClockSourceE  eClkSource;
      OTX_INT32 c;

      DemoLogf("Set Clock Source:\n");
      DemoLogf("0 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to Li0 receive span \n");
      DemoLogf("1 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 receive span \n");
      DemoLogf("2 = OTX_CLOCK_SOURCE_LOCAL_2  - LIs synchronized to Li2 receive span \n");
      DemoLogf("3 = OTX_CLOCK_SOURCE_LOCAL_3  - LIs synchronized to Li3 receive span \n");
      DemoLogf("4 = OTX_CLOCK_SOURCE_LOCAL_4  - LIs synchronized to Li4 receive span \n");
      DemoLogf("5 = OTX_CLOCK_SOURCE_LOCAL_5  - LIs synchronized to Li5 receive span \n");
      DemoLogf("6 = OTX_CLOCK_SOURCE_LOCAL_6  - LIs synchronized to Li6 receive span \n");
      DemoLogf("7 = OTX_CLOCK_SOURCE_LOCAL_7  - LIs synchronized to Li7 receive span \n");
      DemoLogf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");
      
      c = OsaConsoleGetCh();
      if (c == 'i') {
        eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
      } else {
        switch (c - '0') {
        case 0:
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:
        case 7:
          eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_LOCAL_0 + (c - '0') ); 
          break;
        default:
          DemoLogf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          break;
        }
      }
      
      result = OtxBrdSetClocks(g_hPhysDevs.hBoard, OTX_CLOCK_MODE_LOCAL, eClkSource); DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;
  


  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


/*******************************************************************************

  @func

  Enumerates and finds the first board device

*******************************************************************************/
static OTX_RESULT FindBoardDevice(OTX_INT32 *peBrdType) 
{
  OTX_UINT32   eDevType;
  OTX_ITERATOR itrChildTypes;
  OTX_ITERATOR itrChildrenOfSameType;
  static char szBuf[256];
  OTX_RESULT otxRc;

  *peBrdType = OTX_DEVICE_UNKNOWN; // Defensive
  
  // Initialize the opaque iterator for the hosted types
  OtxDrvIteratorBegin(&itrChildTypes);
  while (OTX_SUCCESS(OtxDrvEnumDeviceType(0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {   // Note that we must use a zero handle to get the children of the root level
    OTX_HANDLE hChildDevice;
    // Initialize the opaque iterator for the devices within the same type
    OtxDrvIteratorBegin(&itrChildrenOfSameType);
    while (OTX_SUCCESS(OtxDrvEnumDevice(0, eDevType, &itrChildrenOfSameType, &hChildDevice))) {
      if (hChildDevice != OTX_INVALID_HANDLE_VALUE) {

        // Check device type
        switch (eDevType) {
          // These devices appear at the top (driver level) but they are not board devices, so ignore them here
        case OTX_DEVICE_OTPD_INTERFACE:
        case OTX_DEVICE_CT_BUS:
          break;
        default:
          otxRc = OtxDrvGetAttributeValueSz(hChildDevice, OTX_ATTR_DEVICE_STRING, szBuf, 256);
          if (!OTX_SUCCESS(otxRc)) {
            DemoLogf("Could not get the OTX_ATTR_DEVICE_STRING attribute for device 0x%X\n", hChildDevice);
            return otxRc;
          } 
          DemoLogf("Found %s board (devType=%d)\n", szBuf, eDevType);
          *peBrdType = eDevType;
          return OTX_S_OK;
        }

      }
    }
  }

  return OTX_S_FALSE; // We did not find an OTX board
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  memset(&g_sBurstStat, 0, sizeof(g_sBurstStat));

  DemoLogf("AlvisBurst, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Alvis-CSI ASP Burst Demo program.\n");
  DemoLogf("For Alvis-CSI boards\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  
  f=fopen("out.bin", "wb");

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, 
                                       OTX_ATTR_LIB_REV_STR, 
                                       szVersionInfo, 
                                       255); 
    DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, 
                                         OTX_ATTR_LIB_TYPE_STATIC, 
                                         &bIsStaticLinkedLibrary); 
    DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, 
                                       OTX_ATTR_LIB_FILE_NAME, 
                                       szLibFileName, 
                                       64); 
    DemoCheckRc(result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", 
             szVersionInfo, 
             szLibFileName, 
             bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Query driver revision
  {
    char szVersionInfo[255];
    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_UNKNOWN) {
    // Board type was not supplied by a command line parameter, so enumerate and find a board
    result = FindBoardDevice(&g_demoEnv.nBoardType);
    if (result == OTX_S_FALSE) {
      DemoLogf("No OTX board was found\n");
      DemoExitQuestion();
    } else if (result != OTX_S_OK) {
      DemoExitQuestion();
    }
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
  case OTX_DEVICE_DMP_ALVIS_4_PBX:
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
    // Configure E1/T1 devices
    DemoLogf("Configuring E1/T1 devices...\n");
    DemoConfigureE1T1Devices();
    break;
  default:
    break;
  }

  // Configure socket
#ifdef DEMO_UDP_PACKETS
  {
    struct hostent *hp;
    short nPort = DEMO_DEFAULT_PORT;
    char *strDest = DEMO_DEFAULT_DEST;
    g_addrDest.sin_addr.s_addr = inet_addr((const char*)strDest);
    if(g_addrDest.sin_addr.s_addr == INADDR_NONE) {
      hp = gethostbyname((const char*)strDest);
      if(!hp){
        DemoLogf("base_socket::create ERROR: Failed to resolve host.\n");
        return -1;
      }
      memset(&g_addrDest, 0, sizeof(g_addrDest));
      memcpy(&(g_addrDest.sin_addr), hp->h_addr, hp->h_length);
      g_addrDest.sin_family = hp->h_addrtype;
    } else {
      g_addrDest.sin_family = AF_INET;
    }
    
    g_addrDest.sin_port = htons(nPort);
    
    if (g_hSocket.create_server("0.0.0.0", nPort)) {
      printf("Failed to bind a socket\n"); fflush(stdout);
      return 0;
    } 
    
  }
#endif  

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling Physical Devices...\n");
  DemoEnablePhysicalDevices();

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
  case OTX_DEVICE_DMP_ALVIS_4_PBX:
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
    DemoLogf("Setting clocks...\n");
    result = OtxBrdSetClocks(g_hPhysDevs.hBoard, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(result);
    //result = OtxBrdSetClocks(g_hPhysDevs.hBoard, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(result);
    break;
  default:
    break;
  }


  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
  case OTX_DEVICE_DMP_ALVIS_4_PBX:
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
    // Read Serial Number from the board
    {
      char szSerialNumber[64];
      
      result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoard, OTX_ATTR_BRD_SERIAL_NUMBER, szSerialNumber, 64); DemoCheckRc(result);
      DemoLogf("Board Serial Number: %s\n", szSerialNumber);
    }
    
    // Read Build Number from the board
    {
      char szBrdBuildNumber[32];
      
      result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoard, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 32); DemoCheckRc(result); DemoLogf("Board Build Number: %s\n", szBrdBuildNumber);
    }
    break;
  default:
    break;
  }


  // Main Loop. Loop here and wait for 
  // Events from the driver or commands from the user ....

  OsaThreadStart(DemoBurstIdleFunction);   // This thread runs at a higher priority and handles events from the burst device only.
  OsaThreadStart(DemoIdleFunction);

  //OsaThreadSleep(300);

  DemoLogf("\nPress 'h' for help\n");

#if 1 // Don't use IDLE character
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
  case OTX_DEVICE_DMP_ALVIS_4_PBX:
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
    {
      int i;
      for (i=0; i<g_demoEnv.coLi; i++) {
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[i], 48, 0x00); DemoCheckRc(result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[i], 49, 0x00); DemoCheckRc(result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[i], 50, 0x00); DemoCheckRc(result);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[i], 51, 0x00); DemoCheckRc(result);
      }
    }
    break;
  default:
    break;
  }
#endif

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  DemoPrintf("Exit AlvisBurst Demo\n");
  fclose(f);

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}


